use crate::prelude::*;

use anyhow::Result;

pub(crate) struct CommandPoolBuilder {
    flags: VkCommandPoolCreateFlagBits,
    queue_family_index: u32,
}

impl CommandPoolBuilder {
    pub(crate) fn set_flags(mut self, flags: impl Into<VkCommandPoolCreateFlagBits>) -> Self {
        self.flags = flags.into();

        self
    }

    pub(crate) fn set_queue_family_index(mut self, queue_family_index: u32) -> Self {
        self.queue_family_index = queue_family_index;

        self
    }

    pub(crate) fn build<'a>(self, device: &'a Device) -> Result<CommandPool<'a>> {
        let command_pool_ci = VkCommandPoolCreateInfo::new(self.flags, self.queue_family_index);

        let command_pool = device.create_command_pool(&command_pool_ci)?;

        Ok(CommandPool {
            device,
            command_pool,
        })
    }
}

#[derive(Debug)]
pub(crate) struct CommandPool<'a> {
    device: &'a Device,
    command_pool: VkCommandPool,
}

impl<'a> CommandPool<'a> {
    pub(crate) fn builder() -> CommandPoolBuilder {
        CommandPoolBuilder {
            flags: VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT
                | VK_COMMAND_POOL_CREATE_TRANSIENT_BIT,
            queue_family_index: 0,
        }
    }
}

impl<'a> VulkanDevice for CommandPool<'a> {
    fn device(&self) -> &Device {
        &self.device
    }
}

impl_vk_handle!(CommandPool<'a>, VkCommandPool, command_pool);

impl<'a> Drop for CommandPool<'a> {
    fn drop(&mut self) {
        self.device.destroy_command_pool(self.command_pool);
    }
}
