use crate::prelude::*;

use anyhow::Result;

pub struct DescriptorPoolBuilder<'a> {
    layout: Option<&'a DescriptorSetLayout<'a>>,
    descriptor_count: u32,
    flags: VkDescriptorPoolCreateFlagBits,
}

impl<'a> DescriptorPoolBuilder<'a> {
    pub fn set_flags(mut self, flags: impl Into<VkDescriptorPoolCreateFlagBits>) -> Self {
        self.flags |= flags.into();

        self
    }

    pub fn set_descriptor_set_count(mut self, count: u32) -> Self {
        self.descriptor_count = count;

        self
    }

    pub fn set_layout(mut self, layout: &'a DescriptorSetLayout<'a>) -> Self {
        self.layout = Some(layout);

        self
    }

    pub fn build(self, device: &'a Device) -> Result<DescriptorPool<'a>> {
        if cfg!(debug_assertions) {
            if self.layout.is_none() {
                panic!("no layout set!");
            }

            if self.descriptor_count == 0 {
                panic!("descriptor count must be greater than 0");
            }
        }

        let layout = self.layout.expect("descriptor set layout was not set!");

        let pool_sizes: Vec<VkDescriptorPoolSize> = layout
            .pool_sizes()
            .iter()
            .cloned()
            .map(|mut pool_size| {
                pool_size.descriptorCount *= self.descriptor_count;
                pool_size
            })
            .collect();

        let descriptor_pool_ci = VkDescriptorPoolCreateInfo::new(
            self.flags,
            self.descriptor_count,
            pool_sizes.as_slice(),
        );

        let descriptor_pool = device.create_descriptor_pool(&descriptor_pool_ci)?;

        Ok(DescriptorPool {
            device,
            descriptor_pool,
            descriptor_set_layout: layout,
        })
    }
}

#[derive(Debug)]
pub struct DescriptorPool<'a> {
    device: &'a Device,
    descriptor_pool: VkDescriptorPool,
    pub(crate) descriptor_set_layout: &'a DescriptorSetLayout<'a>,
}

impl<'a> DescriptorPool<'a> {
    pub fn builder() -> DescriptorPoolBuilder<'a> {
        DescriptorPoolBuilder {
            layout: None,
            descriptor_count: 1,
            flags: VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT.into(),
        }
    }

    pub fn reset(&self) -> Result<()> {
        self.device
            .reset_descriptor_pool(self.descriptor_pool, VK_DESCRIPTOR_POOL_RESET_NULL_BIT)
    }

    pub fn prepare_set(&self) -> DescriptorSetBuilder<'a> {
        DescriptorSet::builder(self.device, self)
    }
}

impl<'a> VulkanDevice for DescriptorPool<'a> {
    fn device(&self) -> &Device {
        &self.device
    }
}

impl_vk_handle!(DescriptorPool<'a>, VkDescriptorPool, descriptor_pool);

impl<'a> VkHandle<VkDescriptorSetLayout> for DescriptorPool<'a> {
    fn vk_handle(&self) -> VkDescriptorSetLayout {
        self.descriptor_set_layout.vk_handle()
    }
}

impl<'a> VkHandle<VkDescriptorSetLayout> for &'a DescriptorPool<'a> {
    fn vk_handle(&self) -> VkDescriptorSetLayout {
        self.descriptor_set_layout.vk_handle()
    }
}

impl<'a> Drop for DescriptorPool<'a> {
    fn drop(&mut self) {
        self.device.destroy_descriptor_pool(self.descriptor_pool);
    }
}
