use crate::prelude::*;

use anyhow::Result;

use std::{sync::Arc, time::Duration};

pub struct FenceBuilder {
    signaled: bool,
}

impl FenceBuilder {
    pub fn set_signaled(mut self, signaled: bool) -> Self {
        self.signaled = signaled;

        self
    }

    pub fn build<'a>(self, device: &'a Device) -> Result<Arc<Fence<'a>>> {
        let flag: VkFenceCreateFlagBits = if self.signaled {
            VK_FENCE_CREATE_SIGNALED_BIT.into()
        } else {
            0u32.into()
        };

        let fence_ci = VkFenceCreateInfo::new(flag);

        let fence = device.create_fence(&fence_ci)?;

        Ok(Arc::new(Fence { device, fence }))
    }
}

#[derive(Debug)]
pub struct Fence<'a> {
    device: &'a Device,
    fence: VkFence,
}

impl<'a> Fence<'a> {
    pub fn builder() -> FenceBuilder {
        FenceBuilder { signaled: false }
    }

    pub fn wait(self: &Arc<Self>, timeout: Duration) -> Result<()> {
        self.device.wait_for_fences(&[self], true, timeout)
    }

    pub fn reset(&self) -> bool {
        self.device.reset_fences(&[self.fence]).is_ok()
    }
}

impl<'a> VulkanDevice for Fence<'a> {
    fn device(&self) -> &Device {
        &self.device
    }
}

impl_vk_handle!(Fence<'a>, VkFence, fence);

impl<'a> Drop for Fence<'a> {
    fn drop(&mut self) {
        self.device.destroy_fence(self.fence);
    }
}
