use crate::prelude::*;

use anyhow::Result;

use utilities::impl_reprc;
use vma_rs::prelude::*;

use std::marker::PhantomData;

impl_reprc!(
    #[derive(Debug, Copy)]
    pub struct RawBuffer {
        #[assume_reprc]
        d: u8,
    }
);

#[derive(Debug, Clone, PartialEq, Hash, Eq)]
pub enum MemoryUsage {
    GpuOnly,
    CpuOnly,
    CpuToGpu,
    GpuToCpu,
}

impl MemoryUsage {
    pub fn into_vma(usage: Option<Self>) -> VmaMemoryUsage {
        match usage {
            Some(usage) => usage.into(),
            None => VMA_MEMORY_USAGE_UNKNOWN,
        }
    }
}

impl Into<VmaMemoryUsage> for MemoryUsage {
    fn into(self) -> VmaMemoryUsage {
        match self {
            Self::GpuOnly => VMA_MEMORY_USAGE_GPU_ONLY,
            Self::CpuOnly => VMA_MEMORY_USAGE_CPU_ONLY,
            Self::CpuToGpu => VMA_MEMORY_USAGE_CPU_TO_GPU,
            Self::GpuToCpu => VMA_MEMORY_USAGE_GPU_TO_CPU,
        }
    }
}

#[derive(Debug)]
pub struct Memory<'a, T: ReprC> {
    device: &'a Device,

    allocation: Allocation,

    data_type: PhantomData<T>,
}

impl<'a, T: ReprC> Memory<'a, T> {
    pub(crate) fn forced_requirements(
        device: &'a Device,
        memory_requirements: VkMemoryRequirements,
        buffer: VkBuffer,
        memory_usage: VmaMemoryUsage,
    ) -> Result<Memory<'a, T>> {
        let mut memory = Self::create_and_bind(device, memory_requirements, memory_usage, ())?;
        memory.allocation.bind_buffer_memory(buffer)?;

        Ok(memory)
    }

    pub(crate) fn buffer_memory(
        device: &'a Device,
        buffer: VkBuffer,
        memory_usage: VmaMemoryUsage,
    ) -> Result<Memory<'a, T>> {
        let memory_requirements = device.buffer_memory_requirements(buffer);

        Self::create_and_bind(device, memory_requirements, memory_usage, buffer)
    }

    pub(crate) fn image_memory(
        device: &'a Device,
        image: VkImage,
        memory_usage: VmaMemoryUsage,
    ) -> Result<Memory<'a, T>> {
        let memory_requirements = device.image_memory_requirements(image);

        Self::create_and_bind(device, memory_requirements, memory_usage, image)
    }

    pub(crate) fn vk_handle(&self) -> VkDeviceMemory {
        self.allocation.device_memory()
    }
}

trait MemoryBinder<'a, T, K: ReprC> {
    fn create_and_bind(
        device: &'a Device,
        memory_requirements: VkMemoryRequirements,
        memory_usage: VmaMemoryUsage,
        argument: T,
    ) -> Result<Memory<'a, K>>;
}

impl<'a, K: ReprC> MemoryBinder<'a, (), K> for Memory<'a, K> {
    fn create_and_bind(
        device: &'a Device,
        memory_requirements: VkMemoryRequirements,
        memory_usage: VmaMemoryUsage,
        _: (),
    ) -> Result<Memory<'a, K>> {
        let allocation = device
            .allocator()
            .allocate()
            .set_usage(memory_usage)
            .set_memory_type_bits(memory_requirements.memoryTypeBits.into())
            .build(&memory_requirements)?;

        Ok(Memory {
            device,

            allocation,

            data_type: PhantomData,
        })
    }
}

impl<'a, K: ReprC> MemoryBinder<'a, VkImage, K> for Memory<'a, K> {
    fn create_and_bind(
        device: &'a Device,
        memory_requirements: VkMemoryRequirements,
        memory_usage: VmaMemoryUsage,
        image: VkImage,
    ) -> Result<Memory<'a, K>> {
        let allocation = device
            .allocator()
            .allocate()
            .set_usage(memory_usage)
            .set_memory_type_bits(memory_requirements.memoryTypeBits.into())
            .build(image)?;

        Ok(Memory {
            device,

            allocation,

            data_type: PhantomData,
        })
    }
}

impl<'a, K: ReprC> MemoryBinder<'a, VkBuffer, K> for Memory<'a, K> {
    fn create_and_bind(
        device: &'a Device,
        memory_requirements: VkMemoryRequirements,
        memory_usage: VmaMemoryUsage,
        buffer: VkBuffer,
    ) -> Result<Memory<'a, K>> {
        let allocation = device
            .allocator()
            .allocate()
            .set_usage(memory_usage)
            .set_memory_type_bits(memory_requirements.memoryTypeBits.into())
            .build(buffer)?;

        Ok(Memory {
            device,

            allocation,

            data_type: PhantomData,
        })
    }
}

impl<'a, T: ReprC> VulkanDevice for Memory<'a, T> {
    fn device(&self) -> &Device {
        &self.device
    }
}

impl<'a, T: ReprC + Clone> Memory<'a, T> {
    pub fn map(&self, length: VkDeviceSize) -> Result<VkMappedMemory<'_, T>> {
        self.allocation.map(length)
    }
}
