use anyhow::Result;

use crate::pipeline::PipelineType;
use crate::prelude::*;

pub struct ComputePipelineBuilder<'a> {
    shader_module: Option<&'a ShaderModule<'a, shader_type::Compute>>,
    pipeline_cache: Option<PipelineCache>,
    flags: VkPipelineCreateFlagBits,
}

impl<'a> ComputePipelineBuilder<'a> {
    // TODO: add support for specialization constants
    pub fn set_shader_module(
        mut self,
        shader_module: &'a ShaderModule<'a, shader_type::Compute>,
    ) -> Self {
        if cfg!(debug_assertions) {
            if self.shader_module.is_some() {
                panic!("shader already set!");
            }
        }

        self.shader_module = Some(shader_module);

        self
    }

    pub fn set_pipeline_cache(mut self, pipeline_cache: PipelineCache) -> Self {
        self.pipeline_cache = Some(pipeline_cache);

        self
    }

    pub fn set_flags(mut self, flags: impl Into<VkPipelineCreateFlagBits>) -> Self {
        self.flags = flags.into();

        self
    }

    pub fn build(
        self,
        device: &'a Device,
        pipeline_layout: PipelineLayout,
    ) -> Result<Pipeline<'a>> {
        let pipeline_ci = match self.shader_module {
            Some(module) => VkComputePipelineCreateInfo::new(
                self.flags,
                module.pipeline_stage_info(),
                pipeline_layout.vk_handle(),
            ),
            None => {
                return Err(anyhow::Error::msg(
                    "Required shader module could not be found",
                ))
            }
        };

        let pipeline = device.create_compute_pipelines(
            self.pipeline_cache.map(|cache| cache.vk_handle()),
            &[pipeline_ci],
        )?[0];

        Ok(Pipeline::new(
            device,
            pipeline_layout,
            PipelineType::Compute,
            pipeline,
        ))
    }
}

impl<'a> Default for ComputePipelineBuilder<'a> {
    fn default() -> Self {
        ComputePipelineBuilder {
            shader_module: None,
            pipeline_cache: None,
            flags: 0.into(),
        }
    }
}
