use crate::prelude::*;
use anyhow::Result;

use std::sync::Mutex;
use std::time::Duration;

pub struct SingleSubmit<'a, F, T>
where
    F: FnOnce(&mut CommandBufferRecorder<'_, '_>) -> Result<T>,
{
    command_buffer: &'a mut CommandBuffer<'a>,
    queue: &'a Mutex<Queue<'a>>,
    f: F,

    timeout: Option<Duration>,
}

impl<'a, F, T> SingleSubmit<'a, F, T>
where
    F: FnOnce(&mut CommandBufferRecorder<'_, '_>) -> Result<T>,
{
    pub fn builder(
        command_buffer: &'a mut CommandBuffer<'a>,
        queue: &'a Mutex<Queue<'a>>,
        f: F,
    ) -> Self {
        SingleSubmit {
            command_buffer,
            queue,
            f,

            timeout: None,
        }
    }

    pub fn wait_for_timeout(mut self, timeout: Duration) -> Self {
        self.timeout = Some(timeout);

        self
    }

    pub fn submit(mut self) -> Result<T> {
        let result = {
            let mut buffer_recorder = self.command_buffer.begin(VkCommandBufferBeginInfo::new(
                VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT,
            ))?;

            (self.f)(&mut buffer_recorder)?
        };

        let submit = SubmitInfo::default().add_command_buffer(&*self.command_buffer);
        let queue_lock = self.queue.lock().unwrap();

        match self.timeout {
            Some(timeout) => {
                let fence = Fence::builder().build(self.command_buffer.device())?;

                queue_lock.submit(Some(&fence), &[submit])?;

                self.command_buffer
                    .device()
                    .wait_for_fences(&[&fence], true, timeout)?;
            }
            None => {
                queue_lock.submit(None, &[submit])?;

                queue_lock.wait_idle()?;
            }
        }

        Ok(result)
    }
}
